/* johd: Main Jabber-Over-HTTP daemon.
   Copyright (C) 2011 Sergey Poznyakoff

   JOH is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 3, or (at your option)
   any later version.

   JOH is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with JOH.  If not, see <http://www.gnu.org/licenses/>. */

#include "joh.h"
#include <sys/types.h>
#include <unistd.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <pwd.h>
#include <grp.h>

struct joh_sockaddr *listen_addrs;
struct joh_sockaddr *jabber_addr;
char *pidfile = NULL;
struct joh_http_get_reply *reply;

void
version()
{
	printf("%s (%s) %s\n", program_name, PACKAGE, PACKAGE_VERSION);
	printf("Copyright (C) 2011 Sergey Poznyakoff");
	puts("\
\n\
License GPLv3+: GNU GPL version 3 or later <http://gnu.org/licenses/gpl.html>\n\
This is free software: you are free to change and redistribute it.\n\
There is NO WARRANTY, to the extent permitted by law.\n\
\n\
");
	exit(EX_OK);
}

void
help()
{
	printf("Usage: johd [OPTIONS]\n");
	printf("johd provides a jabber-over-HTTP tunnel (XEP-0025)\n\n");
	printf("OPTIONS:\n");
	puts("\
  -c CLASS         set client class (CGI or HTTP) for subsequent -l options\n\
  -l URL           listen on the given URL (several -l options are OK)\n\
  -s URL           jabber server\n\
  -S NAME          set service name for TCP wrappers\n\
  -t CLASS:TIMEOUT set session idle timeout\n\
  -D               after startup, switch to the background\n\
  -p FILE          write PID to FILE\n\
  -u NAME          after startup, switch to privileges of the user NAME\n\
\n\
HTTP responses:\n\
  -E FILE          Read 404 error page from FILE\n\
  -R URL           Redirect GET requests to URL\n\
\n\
Logging and debugging options:\n\
  -d NUMBER        set debugging level\n\
  -F FACILITY      set syslog facility\n\
  -i               show source line information with debug messages\n\
  -L STRING        set log tag\n\
  -P               prefix diagnostic messages with their severity level\n\
\n\
  -V               print the program version\n\
  -h               show this help summary\n\
\n");
	printf("Report bugs to <%s>.\n", PACKAGE_BUGREPORT);
	exit(EX_OK);
}

static void
set_timeout(char *arg)
{
	enum joh_conntype sesstype;
	time_t timeout;
	int ind = -1;
	static struct { char s[3]; unsigned f; } tab[] = {
		{ "hH", 3600 },
		{ "mM", 60 },
		{ "sS", 1 }
	};
	
	switch (*arg) {
	case 'j':
	case 'J':
		sesstype = conn_jabber;
		break;
	case 'c':
	case 'C':
		sesstype = conn_client;
		break;
	default:
		die(EX_USAGE, "bad type in timeout specification: %s", arg);
	}
	if (*++arg != ':')
		die(EX_USAGE, "timeout specification missing colon: %s", arg);
	timeout = 0;
	for (;;) {
		unsigned long t = strtoul(arg + 1, &arg, 10);
		if (*arg) {
			unsigned x = 0;
			
			for (ind++; ind<sizeof(tab)/sizeof(tab[0]);
			     ind++) {
				if (strchr(tab[ind].s, *arg)) {
					x = tab[ind].f;
					break;
				}
			}
			if (x == 0)
				die(EX_USAGE, "unknown time suffix %c", *arg);
			timeout += t * x;
			if (x > 1)
				continue;
		}
		break;
	}
	if (*arg)
		die(EX_USAGE, "invalid timeout specification: %s", arg);
	joh_session_set_timeout(sesstype, timeout);
}

void
joh_check_pidfile(const char *name)
{
	unsigned long pid;
	FILE *fp = fopen(name, "r");
	if (!fp) {
		if (errno == ENOENT)
			return;
		error("cannot open pidfile `%s': %s",
		      name, strerror(errno));
		exit(EX_TEMPFAIL);
	}
	if (fscanf(fp, "%lu", &pid) != 1) {
		error("cannot read pidfile `%s'", name);
	} else {
		if (kill(pid, 0) == 0) {
			error("%s appears to run with pid %lu; "
			      "if it does not, remove `%s' and retry",
				program_name,
				pid,
			      name);
			exit(EX_USAGE);
		}
	}
	fclose(fp);
	if (unlink(name)) {
		error("cannot unlink pidfile `%s': %s",
		      name, strerror(errno));
		exit(EX_USAGE);
	}
}

void
joh_remove_pidfile(void)
{
	unlink(pidfile);
}

void
joh_write_pidfile(const char *name)
{
	FILE *fp = fopen(name, "w");
	if (!fp) {
		error("cannot create pidfile `%s': %s",
		      name, strerror(errno));
		exit(EX_TEMPFAIL);
	}
	fprintf(fp, "%lu\n", (unsigned long) getpid());
	fclose(fp);
	atexit(joh_remove_pidfile);
}

static void
new_get_reply(struct joh_sockaddr_hints *hints, int redirect, const char *text)
{
	char *data;
	size_t length;

	if (redirect) {
		if (!strprefix(text, "http://"))
			die(EX_USAGE,
			    "argument to -R must begin with http://");
		data = (char*)text;
		length = strlen(text);
	} else {
		struct stat st;
		int fd;
		ssize_t rc;

		if (stat(text, &st))
			die(EX_USAGE,
			    "cannot open error file %s: %s",
			    text, strerror(errno));
		length = st.st_size;
		if (length != st.st_size)
			die(EX_USAGE,
			    "file %s is too big", text);
		data = xmalloc(length + 1);
		fd = open(text, O_RDONLY);
		if (fd == -1)
			die(EX_UNAVAILABLE,
			    "cannot open error file %s: %s",
			    text, strerror(errno));
		rc = read(fd, data, length);
 		if (rc != length)
			die(EX_UNAVAILABLE,
			    "error reading file %s: %s",
			    text, strerror(errno));
		/* FIXME: translate LF to CRLF? */
		close(fd);
	}
	
	hints->reply = xmalloc(sizeof(hints->reply[0]));
	hints->reply->redirect = redirect;
	hints->reply->data = data;
	hints->reply->length = length;
	hints->flags |= JOH_HINT_REPLY;
}

static void
setuserpriv(const char *user_name)
{
	struct passwd *pw;
	
	if (!user_name)
		return;
	pw = getpwnam(user_name);
	if (!pw)
		die(EX_USAGE, "no such user: %s", user_name);
	debug(2, ("switching to privileges of %s (%lu:%lu)",
		  user_name, (unsigned long)pw->pw_uid,
		  (unsigned long)pw->pw_gid));
	if (setgroups(1, &pw->pw_gid))
		die(EX_UNAVAILABLE,
		    "setgroups failed: %s", strerror(errno));
	if (setgid(pw->pw_uid))
		die(EX_UNAVAILABLE,
		    "setgid(%lu) failed: %s",
		    (unsigned long) pw->pw_gid, strerror(errno));
	if (setuid(pw->pw_uid))
		die(EX_UNAVAILABLE,
		    "setuid(%lu) failed: %s",
		    (unsigned long) pw->pw_uid, strerror(errno));
}

int
main(int argc, char **argv)
{
	int c;
	char *p;
	int daemon_option = 0;
	char *user_name = NULL;
	struct joh_sockaddr *jsa;
	static struct joh_sockaddr_hints
		listen_hints = { JOH_HINT_BIND|JOH_HINT_SERVICE|JOH_HINT_CLASS,
				 DEFAULT_PORT_STRING, client_cgi },
		jabber_hints = { JOH_HINT_SERVICE, JABBER_PORT_STRING };
	     
	set_program_name(argv[0]);
	while ((c = getopt(argc, argv, "c:d:DE:F:il:L:s:S:p:PR:t:u:hV"))
	       != EOF) {
		switch (c) {
		case 'c':
			if (strcasecmp(optarg, "HTTP") == 0) {
				listen_hints.class = client_http;
				listen_hints.service = "80";
			} else if (strcasecmp(optarg, "CGI") == 0) {
				listen_hints.class = client_cgi;
				listen_hints.service = DEFAULT_PORT_STRING;
			} else
				die(EX_USAGE,
				    "unrecognized client class: %s", optarg);
			break;
			
		case 'd':
			debug_level = strtoul(optarg, &p, 10);
			if (*p)
				die(EX_USAGE,
				    "debug level must be a number");
			else if (debug_level < 0 || debug_level > 100)
				die(EX_USAGE,
				    "debug level out of range 0..100");
			break;

		case 'D':
			daemon_option = 1;
			break;

		case 'E':
			new_get_reply(&listen_hints, 0, optarg);
			break;

		case 'F':
			facility_option = optarg;
			break;
			
		case 'i':
			source_info_option = 1;
			break;
			
		case 'l':
			jsa = parse_sockaddr(optarg, &listen_hints);
			listen_addrs =
				joh_sockaddr_append(listen_addrs, jsa);
			break;

		case 'L':
			log_tag = optarg;
			break;
			
		case 'P':
			prefix_severity_option = 1;
			break;

		case 'p':
			pidfile = optarg;
			break;
			
		case 'R':
			new_get_reply(&listen_hints, 1, optarg);
			break;
			
		case 's':
			if (jabber_addr)
				die(EX_USAGE,
				    "only one -s option is allowed");
			jabber_addr = parse_sockaddr(optarg,
						     &jabber_hints);
			if (jabber_addr->next) {
				warn("%s resolves to more than one address, "
				     "using %s",
				     optarg, joh_sockaddr_str(jabber_addr));
				joh_sockaddr_free(jabber_addr->next);
				jabber_addr->next = NULL;
			}
			break;

		case 'S':
			listen_hints.flags |= JOH_HINT_SRVNAME;
			listen_hints.srvname = optarg;
			break;

		case 't':
			set_timeout(optarg);
			break;

		case 'u':
			user_name = optarg;
			break;
			
		case 'h':
			help();
			break;

		case 'V':
			version();
			break;

		default:
			exit(EX_USAGE);
		}
	}

	if (optind < argc)
		die(EX_USAGE, "too many arguments");

	if (pidfile)
		joh_check_pidfile(pidfile);
	
	if (daemon_option) {
		if (facility_option)
			facility_option = "daemon";
		if (daemon(0, 0))
			die(EX_UNAVAILABLE,
			    "cannot switch to background: %s",
			    strerror(errno));
	}
	
	joh_initlog();
	debug(1, ("johd %s startup", PACKAGE_VERSION));
	if (!listen_addrs)
		listen_addrs = parse_sockaddr("inet://", &listen_hints);
	if (!jabber_addr) {
		/* FIXME: Use SRV record, if available. Ditto for -l. */
		jabber_addr = parse_sockaddr(JABBER_SERVER, &jabber_hints);
		if (jabber_addr->next) {
			joh_sockaddr_free(jabber_addr->next);
			jabber_addr->next = NULL;
		}
	}

	joh_conntab_init();
	joh_listen_socket_setup(listen_addrs);
	setuserpriv(user_name);
	if (pidfile)
		joh_write_pidfile(pidfile);
	joh_loop();
	joh_listen_socket_cleanup();
	debug(1, ("johd %s terminated", PACKAGE_VERSION));
	
	return EX_OK;
}
